@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2019 Gentoo Foundation
+ * Copyright 2005-2020 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2008 Ned Ludd        - <solar@gentoo.org>
@@ -27,6 +27,8 @@
 #include <xalloc.h>
 
 static int tree_pkg_compar(const void *l, const void *r);
+static tree_pkg_ctx * tree_next_pkg_int(tree_cat_ctx *cat_ctx);
+static void tree_close_meta(tree_pkg_meta *cache);
 
 static tree_ctx *
 tree_open_int(const char *sroot, const char *tdir, bool quiet)
@@ -57,8 +59,6 @@ tree_open_int(const char *sroot, const char *tdir, bool quiet)
 		goto cv_error;
 
 	ctx->do_sort = false;
-	ctx->catsortfunc = alphasort;
-	ctx->pkgsortfunc = tree_pkg_compar;
 	return ctx;
 
  cv_error:
@@ -163,7 +163,7 @@ tree_close(tree_ctx *ctx)
 	free(ctx);
 }
 
-int
+static int
 tree_filter_cat(const struct dirent *de)
 {
 	int i;
@@ -232,28 +232,36 @@ tree_open_cat(tree_ctx *ctx, const char *name)
 	return cat_ctx;
 }
 
-tree_cat_ctx *
+static tree_cat_ctx *
 tree_next_cat(tree_ctx *ctx)
 {
 	/* search for a category directory */
 	tree_cat_ctx *cat_ctx = NULL;
+	const struct dirent *de;
 
 	if (ctx->do_sort) {
 		if (ctx->cat_de == NULL) {
 			ctx->cat_cnt = scandirat(ctx->tree_fd,
-					".", &ctx->cat_de, tree_filter_cat, ctx->catsortfunc);
+					".", &ctx->cat_de, tree_filter_cat, alphasort);
 			ctx->cat_cur = 0;
 		}
 
 		while (ctx->cat_cur < ctx->cat_cnt) {
-			cat_ctx = tree_open_cat(ctx, ctx->cat_de[ctx->cat_cur++]->d_name);
+			de = ctx->cat_de[ctx->cat_cur++];
+
+			/* match if cat is requested */
+			if (ctx->query_atom != NULL && ctx->query_atom->CATEGORY != NULL &&
+					strcmp(ctx->query_atom->CATEGORY, de->d_name) != 0)
+				continue;
+
+			cat_ctx = tree_open_cat(ctx, de->d_name);
 			if (!cat_ctx)
 				continue;
+
 			break;
 		}
 	} else {
 		/* cheaper "streaming" variant */
-		const struct dirent *de;
 		do {
 			de = readdir(ctx->dir);
 			if (!de)
@@ -262,6 +270,11 @@ tree_next_cat(tree_ctx *ctx)
 			if (tree_filter_cat(de) == 0)
 				continue;
 
+			/* match if cat is requested */
+			if (ctx->query_atom != NULL && ctx->query_atom->CATEGORY != NULL &&
+					strcmp(ctx->query_atom->CATEGORY, de->d_name) != 0)
+				continue;
+
 			cat_ctx = tree_open_cat(ctx, de->d_name);
 			if (!cat_ctx)
 				continue;
@@ -284,7 +297,7 @@ tree_close_cat(tree_cat_ctx *cat_ctx)
 	free(cat_ctx);
 }
 
-int
+static int
 tree_filter_pkg(const struct dirent *de)
 {
 	int i;
@@ -326,6 +339,18 @@ tree_open_pkg(tree_cat_ctx *cat_ctx, const char *name)
 	pkg_ctx->fd = -1;
 	pkg_ctx->cat_ctx = cat_ctx;
 	pkg_ctx->atom = NULL;
+	pkg_ctx->meta = NULL;
+
+	/* see if this pkg matches the query, here we can finally check
+	 * version conditions like >=, etc. */
+	if (cat_ctx->ctx->query_atom != NULL) {
+		(void)tree_get_atom(pkg_ctx, false);
+		if (atom_compare(pkg_ctx->atom, cat_ctx->ctx->query_atom) != EQUAL) {
+			tree_close_pkg(pkg_ctx);
+			return NULL;
+		}
+	}
+
 	return pkg_ctx;
 }
 
@@ -340,13 +365,12 @@ tree_pkg_compar(const void *l, const void *r)
 	return atom_compar_cb(al, ar);
 }
 
-static tree_pkg_ctx *
-tree_next_pkg_int(tree_cat_ctx *cat_ctx);
 static tree_pkg_ctx *
 tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 {
 	tree_pkg_ctx *pkg_ctx = NULL;
 	const struct dirent *de;
+	depend_atom *qa = cat_ctx->ctx->query_atom;
 
 	if (cat_ctx->ctx->do_sort) {
 		if (cat_ctx->pkg_ctxs == NULL) {
@@ -359,11 +383,22 @@ tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 				if (tree_filter_pkg(de) == 0)
 					continue;
 
+				/* perform package name check, for we don't have an atom
+				 * yet, and creating it is expensive, which we better
+				 * defer to pkg time, and filter most stuff out here
+				 * note that we might over-match, but that's easier than
+				 * trying to deal with end of string or '-' here (which
+				 * still wouldn't be 100% because name rules are complex) */
+				if (qa != NULL && qa->PN != NULL &&
+						strncmp(qa->PN, de->d_name, strlen(qa->PN)) != 0)
+					continue;
+
 				if (cat_ctx->pkg_cnt == pkg_size) {
 					pkg_size += 256;
 					cat_ctx->pkg_ctxs = xrealloc(cat_ctx->pkg_ctxs,
 								sizeof(*cat_ctx->pkg_ctxs) * pkg_size);
 				}
+
 				name = xstrdup(de->d_name);
 				pkg_ctx = cat_ctx->pkg_ctxs[cat_ctx->pkg_cnt++] =
 					tree_open_pkg(cat_ctx, name);
@@ -373,9 +408,9 @@ tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 				}
 			}
 
-			if (cat_ctx->ctx->pkgsortfunc != NULL && cat_ctx->pkg_cnt > 1) {
+			if (cat_ctx->pkg_cnt > 1) {
 				qsort(cat_ctx->pkg_ctxs, cat_ctx->pkg_cnt,
-						sizeof(*cat_ctx->pkg_ctxs), cat_ctx->ctx->pkgsortfunc);
+						sizeof(*cat_ctx->pkg_ctxs), tree_pkg_compar);
 			}
 		}
 
@@ -391,6 +426,11 @@ tree_next_pkg_int(tree_cat_ctx *cat_ctx)
 			if (tree_filter_pkg(de) == 0)
 				continue;
 
+			/* perform package name check as for the sorted variant */
+			if (qa != NULL && qa->PN != NULL &&
+					strncmp(qa->PN, de->d_name, strlen(qa->PN)) != 0)
+				continue;
+
 			pkg_ctx = tree_open_pkg(cat_ctx, de->d_name);
 			if (!pkg_ctx)
 				continue;
@@ -418,8 +458,7 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 				tree_ctx *pkgdir = ctx->ebuilddir_ctx;
 
 				if (pkgdir == NULL)
-					pkgdir = ctx->ebuilddir_ctx = xmalloc(sizeof(tree_ctx));
-				memset(ctx->ebuilddir_ctx, '\0', sizeof(*ctx->ebuilddir_ctx));
+					pkgdir = ctx->ebuilddir_ctx = xzalloc(sizeof(tree_ctx));
 
 				ctx->ebuilddir_pkg_ctx = tree_next_pkg_int(cat_ctx);
 				if (ctx->ebuilddir_pkg_ctx == NULL)
@@ -428,8 +467,6 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 				pkgdir->portroot_fd = -1;
 				pkgdir->tree_fd = cat_ctx->fd;
 				pkgdir->do_sort = ctx->do_sort;
-				pkgdir->catsortfunc = ctx->catsortfunc;
-				pkgdir->pkgsortfunc = ctx->pkgsortfunc;
 				pkgdir->repo = ctx->repo;
 				pkgdir->cachetype = ctx->cachetype;
 
@@ -440,7 +477,7 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 				 * directory or something */
 				if (ctx->ebuilddir_cat_ctx == NULL) {
 					ctx->ebuilddir_pkg_ctx = NULL;
-					return NULL;
+					continue;
 				}
 
 				/* "zap" the pkg such that it looks like CAT/P */
@@ -474,7 +511,7 @@ tree_next_pkg(tree_cat_ctx *cat_ctx)
 	return ret;
 }
 
-int
+static int
 tree_pkg_vdb_openat(
 		tree_pkg_ctx *pkg_ctx,
 		const char *file,
@@ -490,29 +527,7 @@ tree_pkg_vdb_openat(
 	return openat(pkg_ctx->fd, file, flags | O_CLOEXEC, mode);
 }
 
-FILE *
-tree_pkg_vdb_fopenat(
-		tree_pkg_ctx *pkg_ctx,
-		const char *file,
-		int flags,
-		mode_t mode,
-		const char *fmode)
-{
-	FILE *fp;
-	int fd;
-
-	fd = tree_pkg_vdb_openat(pkg_ctx, file, flags, mode);
-	if (fd == -1)
-		return NULL;
-
-	fp = fdopen(fd, fmode);
-	if (!fp)
-		close(fd);
-
-	return fp;
-}
-
-bool
+static bool
 tree_pkg_vdb_eat(
 		tree_pkg_ctx *pkg_ctx,
 		const char *file,
@@ -554,13 +569,13 @@ tree_read_file_pms(tree_pkg_ctx *pkg_ctx)
 	if ((off_t)fread(ptr, 1, s.st_size, f) != s.st_size)
 		goto err;
 
-	ret->DEPEND = ptr;
+	ret->Q_DEPEND = ptr;
 #define next_line(curr, next) \
-	if ((ptr = strchr(ret->curr, '\n')) == NULL) { \
+	if ((ptr = strchr(ret->Q_##curr, '\n')) == NULL) { \
 		warn("Invalid cache file for '%s'", buf); \
 		goto err; \
 	} \
-	ret->next = ptr+1; \
+	ret->Q_##next = ptr+1; \
 	*ptr = '\0';
 	next_line(DEPEND, RDEPEND)
 	next_line(RDEPEND, SLOT)
@@ -630,7 +645,7 @@ tree_read_file_md5(tree_pkg_ctx *pkg_ctx)
 	 * */
 #define assign_var_cmp(keyname, cmpkey) \
 	if (strncmp(keyptr, cmpkey, strlen(cmpkey)) == 0) { \
-		ret->keyname = valptr; \
+		ret->Q_##keyname = valptr; \
 		continue; \
 	}
 #define assign_var(keyname) \
@@ -684,6 +699,7 @@ tree_read_file_md5(tree_pkg_ctx *pkg_ctx)
 		assign_var(DEFINED_PHASES);
 		assign_var(REQUIRED_USE);
 		assign_var(BDEPEND);
+		assign_var(EPREFIX);
 		assign_var(_eclasses_);
 		assign_var(_md5_);
 		warn("Cache file for '%s/%s' has unknown key %s",
@@ -741,7 +757,7 @@ tree_read_file_ebuild(tree_pkg_ctx *pkg_ctx)
 		if (q < p && *p == '=') {
 			*p++ = '\0';
 			/* match variable against which ones we look for */
-#define match_key(X) else if (strcmp(q, #X) == 0) key = &ret->X
+#define match_key(X) else if (strcmp(q, #X) == 0) key = &ret->Q_##X
 			if (1 == 0); /* dummy for syntax */
 			match_key(DEPEND);
 			match_key(RDEPEND);
@@ -757,6 +773,7 @@ tree_read_file_ebuild(tree_pkg_ctx *pkg_ctx)
 			match_key(PDEPEND);
 			match_key(EAPI);
 			match_key(REQUIRED_USE);
+			match_key(BDEPEND);
 #undef match_key
 		}
 
@@ -840,7 +857,7 @@ tree_read_file_binpkg_xpak_cb(
 
 #define match_path(K) \
 	else if (pathname_len == (sizeof(#K) - 1) && strcmp(pathname, #K) == 0) \
-		key = &m->K
+		key = &m->Q_##K
 	if (1 == 0); /* dummy for syntax */
 	match_path(DEPEND);
 	match_path(RDEPEND);
@@ -862,29 +879,40 @@ tree_read_file_binpkg_xpak_cb(
 	match_path(BDEPEND);
 	match_path(CONTENTS);
 	match_path(USE);
+	match_path(EPREFIX);
 	match_path(repository);
 	else
 		return;
 #undef match_path
 
 	/* hijack unused members */
-	pos = (size_t)m->_eclasses_;
-	len = (size_t)m->_md5_;
+	pos = (size_t)m->Q__eclasses_;
+	len = (size_t)m->Q__md5_;
 
 	/* trim whitespace (mostly trailing newline) */
 	while (isspace((int)data[data_offset + data_len - 1]))
 		data_len--;
 
-	if (len - pos < (size_t)data_len) {
+	if (len - pos < (size_t)(data_len + 1)) {
+		char *old_data = m->Q__data;
 		len += (((data_len + 1) / BUFSIZ) + 1) * BUFSIZ;
-		m->_data = xrealloc(m->_data, len);
-		m->_md5_ = (char *)len;
+		m->Q__data = xrealloc(m->Q__data, len);
+		m->Q__md5_ = (char *)len;
+
+		/* re-position existing keys */
+		if (old_data != NULL && m->Q__data != old_data) {
+			char **newdata = (char **)m;
+			int elems = sizeof(tree_pkg_meta) / sizeof(char *);
+			while (elems-- > 1)  /* skip Q__data itself */
+				if (newdata[elems] != NULL)
+					newdata[elems] = m->Q__data + (newdata[elems] - old_data);
+		}
 	}
 
-	*key = m->_data + pos;
+	*key = m->Q__data + pos;
 	snprintf(*key, len - pos, "%.*s", data_len, data + data_offset);
 	pos += data_len + 1;
-	m->_eclasses_ = (char *)pos;
+	m->Q__eclasses_ = (char *)pos;
 }
 
 static tree_pkg_meta *
@@ -898,7 +926,7 @@ tree_read_file_binpkg(tree_pkg_ctx *pkg_ctx)
 	return m;
 }
 
-tree_pkg_meta *
+static tree_pkg_meta *
 tree_pkg_read(tree_pkg_ctx *pkg_ctx)
 {
 	tree_ctx *ctx = pkg_ctx->cat_ctx->ctx;
@@ -931,20 +959,111 @@ tree_pkg_read(tree_pkg_ctx *pkg_ctx)
 		return (tree_pkg_meta *)pkg_ctx->cat_ctx->ctx->pkgs;
 	}
 
-	warn("Unknown metadata cache type!");
+	warn("Unknown/unsupported metadata cache type!");
 	return NULL;
 }
 
-void
+static void
 tree_close_meta(tree_pkg_meta *cache)
 {
 	if (cache == NULL)
 		errf("Cache is empty !");
-	if (cache->_data != NULL)
-		free(cache->_data);
+	if (cache->Q__data != NULL)
+		free(cache->Q__data);
 	free(cache);
 }
 
+char *
+tree_pkg_meta_get_int(tree_pkg_ctx *pkg_ctx, size_t offset, const char *keyn)
+{
+	tree_ctx *ctx = pkg_ctx->cat_ctx->ctx;
+	char **key;
+
+	/* offset is a byte offset in the tree_pkg_meta struct, pointing to
+	 * key, the tree_pkg_meta_get macro as called by the user takes care
+	 * of offset and keyn pointing to the same thing */
+
+	if (ctx->cachetype == CACHE_VDB) {
+		if (pkg_ctx->meta == NULL)
+			pkg_ctx->meta = xzalloc(sizeof(tree_pkg_meta));
+
+		key = (char **)((char *)&pkg_ctx->meta->Q__data + offset);
+
+		/* just eat the file if we haven't yet */
+		if (*key == NULL) {
+			int fd = tree_pkg_vdb_openat(pkg_ctx, keyn, O_RDONLY, 0);
+			struct stat s;
+			size_t pos;
+			size_t len;
+			char *p;
+			tree_pkg_meta *m = pkg_ctx->meta;
+
+			if (fd < 0)
+				return NULL;
+			if (fstat(fd, &s) != 0) {
+				close(fd);
+				return NULL;
+			}
+
+			/* hijack unused members */
+			pos = (size_t)m->Q__eclasses_;
+			len = (size_t)m->Q__md5_;
+
+			/* TODO: this is an exact copy from tree_read_file_binpkg_xpak_cb */
+			if (len - pos < (size_t)(s.st_size + 1)) {
+				p = m->Q__data;
+				len += (((s.st_size + 1) / BUFSIZ) + 1) * BUFSIZ;
+				m->Q__data = xrealloc(m->Q__data, len);
+				m->Q__md5_ = (char *)len;
+
+				/* re-position existing keys */
+				if (p != NULL && m->Q__data != p) {
+					char **newdata = (char **)m;
+					int elems = sizeof(tree_pkg_meta) / sizeof(char *);
+					while (elems-- > 1)  /* skip Q__data itself */
+						if (newdata[elems] != NULL)
+							newdata[elems] = m->Q__data + (newdata[elems] - p);
+				}
+			}
+
+			p = *key = m->Q__data + pos;
+			if (read(fd, p, s.st_size) == (ssize_t)s.st_size) {
+				p[s.st_size] = '\0';
+				while (s.st_size > 0 && isspace((int)p[s.st_size - 1]))
+					p[--s.st_size] = '\0';
+				pos += s.st_size + 1;
+				m->Q__eclasses_ = (char *)pos;
+			}
+			close(fd);
+		}
+	} else {
+		if (pkg_ctx->meta == NULL)
+			pkg_ctx->meta = tree_pkg_read(pkg_ctx);
+		if (pkg_ctx->meta == NULL)
+			return NULL;
+
+		key = (char **)((char *)&pkg_ctx->meta->Q__data + offset);
+
+		/* Packages are nice, but also a bit daft, because they don't
+		 * contain everything available (for a semi-good reason though)
+		 * We cannot downgrade the tree execution to BINPKGS, because
+		 * we're running from tree_foreach_packages */
+		if (*key == NULL && ctx->cachetype == CACHE_PACKAGES) {
+			ctx->cachetype = CACHE_BINPKGS;
+			pkg_ctx->fd = -1;
+			pkg_ctx->meta = tree_pkg_read(pkg_ctx);
+			ctx->cachetype = CACHE_PACKAGES;
+			if (pkg_ctx->meta == NULL) {
+				/* hrmffff. */
+				pkg_ctx->fd = 0;
+				pkg_ctx->meta = tree_pkg_read(pkg_ctx);
+			}
+			key = (char **)((char *)&pkg_ctx->meta->Q__data + offset);
+		}
+	}
+	return *key;
+}
+
 tree_metadata_xml *
 tree_pkg_metadata(tree_pkg_ctx *pkg_ctx)
 {
@@ -1048,6 +1167,9 @@ tree_close_pkg(tree_pkg_ctx *pkg_ctx)
 	if (pkg_ctx->cat_ctx->ctx->do_sort)
 		free((char *)pkg_ctx->name);
 	free(pkg_ctx->slot);
+	if (pkg_ctx->meta != NULL &&
+			(void *)pkg_ctx->meta != (void *)pkg_ctx->cat_ctx->ctx->pkgs)
+		tree_close_meta(pkg_ctx->meta);
 	free(pkg_ctx);
 }
 
@@ -1057,18 +1179,17 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 	char *p = ctx->pkgs;
 	char *q;
 	char *c;
+	char pkgname[_Q_PATH_MAX];
 	size_t len = ctx->pkgslen;
 	int ret = 0;
+	depend_atom *query = ctx->query_atom;
 
 	/* reused for every entry */
-	tree_cat_ctx *cat = xzalloc(sizeof(tree_cat_ctx));
+	tree_cat_ctx *cat = NULL;
 	tree_pkg_ctx *pkg = xzalloc(sizeof(tree_pkg_ctx));
 	tree_pkg_meta *meta = xzalloc(sizeof(tree_pkg_meta));
 	depend_atom *atom = NULL;
 
-	cat->ctx = ctx;
-	pkg->cat_ctx = cat;
-
 	do {
 		/* find next line */
 		c = NULL;
@@ -1083,15 +1204,26 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 		if (p == q) {
 			/* make callback with populated atom */
 			if (atom != NULL) {
+				size_t pkgnamelen;
+
 				/* store meta ptr in repo->pkgs, such that get_pkg_meta
 				 * can grab it from there (for free) */
 				ctx->pkgs = (char *)meta;
 
-				cat->name = atom->CATEGORY;
-				pkg->name = atom->PN;
-				pkg->slot = meta->SLOT == NULL ? (char *)"0" : meta->SLOT;
+				if (cat == NULL || strcmp(cat->name, atom->CATEGORY) != 0)
+				{
+					if (cat != NULL)
+						tree_close_cat(cat);
+					pkg->cat_ctx = cat = tree_open_cat(ctx, atom->CATEGORY);
+				}
+				pkgnamelen = snprintf(pkgname, sizeof(pkgname),
+						"%s.tbz2", atom->PF);
+				pkgname[pkgnamelen - (sizeof(".tbz2") - 1)] = '\0';
+				pkg->name = pkgname;
+				pkg->slot = meta->Q_SLOT == NULL ? (char *)"0" : meta->Q_SLOT;
 				pkg->repo = ctx->repo;
 				pkg->atom = atom;
+				pkg->fd = 0;  /* intentional, meta has already been read */
 
 				/* do call callback with pkg_atom (populate cat and pkg) */
 				ret |= callback(pkg, priv);
@@ -1127,10 +1259,15 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 			if (atom != NULL)
 				atom_implode(atom);
 			atom = atom_explode(c);
+			/* pretend this entry is bogus if it doesn't match query */
+			if (query != NULL && atom_compare(atom, query) != EQUAL) {
+				atom_implode(atom);
+				atom = NULL;
+			}
 #define match_key(X) match_key2(X,X)
 #define match_key2(X,Y) \
 		} else if (strcmp(p, #X) == 0) { \
-			meta->Y = c
+			meta->Q_##Y = c
 		match_key(DEFINED_PHASES);
 		match_key(DEPEND);
 		match_key2(DESC, DESCRIPTION);
@@ -1158,9 +1295,8 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 }
 
 int
-tree_foreach_pkg(tree_ctx *ctx,
-		tree_pkg_cb callback, void *priv, tree_cat_filter filter,
-		bool sort, void *catsortfunc, void *pkgsortfunc)
+tree_foreach_pkg(tree_ctx *ctx, tree_pkg_cb callback, void *priv,
+		bool sort, depend_atom *query)
 {
 	tree_cat_ctx *cat_ctx;
 	tree_pkg_ctx *pkg_ctx;
@@ -1169,20 +1305,15 @@ tree_foreach_pkg(tree_ctx *ctx,
 	if (ctx == NULL)
 		return EXIT_FAILURE;
 
+	ctx->do_sort = sort;
+	ctx->query_atom = query;
+
 	/* handle Packages (binpkgs index) file separately */
 	if (ctx->cachetype == CACHE_PACKAGES)
 		return tree_foreach_packages(ctx, callback, priv);
 
-	ctx->do_sort = sort;
-	if (catsortfunc != NULL)
-		ctx->catsortfunc = catsortfunc;
-	if (pkgsortfunc != NULL)
-		ctx->pkgsortfunc = pkgsortfunc;
-
 	ret = 0;
 	while ((cat_ctx = tree_next_cat(ctx))) {
-		if (filter && !filter(cat_ctx, priv))
-			continue;
 		while ((pkg_ctx = tree_next_pkg(cat_ctx))) {
 			ret |= callback(pkg_ctx, priv);
 			tree_close_pkg(pkg_ctx);
@@ -1190,6 +1321,14 @@ tree_foreach_pkg(tree_ctx *ctx,
 		tree_close_cat(cat_ctx);
 	}
 
+	/* allow foreach to be called again on the same open tree */
+	if (ctx->do_sort)
+		scandir_free(ctx->cat_de, ctx->cat_cnt);
+	ctx->cat_de = NULL;
+	ctx->cat_cur = 0;
+	ctx->cat_cnt = 0;
+	ctx->do_sort = 0;
+
 	return ret;
 }
 
@@ -1207,6 +1346,7 @@ tree_get_atom(tree_pkg_ctx *pkg_ctx, bool complete)
 		tree_ctx *ctx = pkg_ctx->cat_ctx->ctx;
 		if (ctx->cachetype == CACHE_VDB) {
 			if (pkg_ctx->atom->SLOT == NULL) {
+				/* FIXME: use tree_meta_get !!! */
 				if (pkg_ctx->slot == NULL)
 					tree_pkg_vdb_eat(pkg_ctx, "SLOT",
 							&pkg_ctx->slot, &pkg_ctx->slot_len);
@@ -1224,8 +1364,8 @@ tree_get_atom(tree_pkg_ctx *pkg_ctx, bool complete)
 				if (pkg_ctx->slot == NULL) {
 					meta = tree_pkg_read(pkg_ctx);
 					if (meta != NULL) {
-						if (meta->SLOT != NULL) {
-							pkg_ctx->slot = xstrdup(meta->SLOT);
+						if (meta->Q_SLOT != NULL) {
+							pkg_ctx->slot = xstrdup(meta->Q_SLOT);
 							pkg_ctx->slot_len = strlen(pkg_ctx->slot);
 						}
 					}
@@ -1237,8 +1377,8 @@ tree_get_atom(tree_pkg_ctx *pkg_ctx, bool complete)
 				if (pkg_ctx->repo == NULL && ctx->cachetype == CACHE_BINPKGS) {
 					if (meta == NULL)
 						meta = tree_pkg_read(pkg_ctx);
-					if (meta != NULL && meta->repository != NULL) {
-						pkg_ctx->repo = xstrdup(meta->repository);
+					if (meta != NULL && meta->Q_repository != NULL) {
+						pkg_ctx->repo = xstrdup(meta->Q_repository);
 						pkg_ctx->repo_len = strlen(pkg_ctx->repo);
 					}
 				}
@@ -1277,9 +1417,7 @@ static int tree_get_atoms_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 	char abuf[BUFSIZ];
 
 	if (state->fullcpv) {
-		snprintf(abuf, sizeof(abuf), "%s/%s-%s",
-				atom->CATEGORY, atom->PN,
-				atom->PR_int > 0 ? atom->PVR : atom->PV);
+		snprintf(abuf, sizeof(abuf), "%s/%s", atom->CATEGORY, atom->PF);
 		state->cpf = add_set(abuf, state->cpf);
 	} else {
 		snprintf(abuf, sizeof(abuf), "%s/%s", atom->CATEGORY, atom->PN);
