@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2019 Gentoo Foundation
+ * Copyright 2005-2020 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2006      Thomas A. Cort - <tcort@gentoo.org>
@@ -169,7 +169,9 @@ print_keywords(const char *category, const char *ebuild, int *keywords)
 static int
 read_keywords(char *s, int *keywords)
 {
-	char *arch, delim[2] = { ' ', '\0' };
+	char *arch;
+	char delim[2] = { ' ', '\0' };
+	char *savep;
 	size_t slen;
 	size_t a;
 	int i;
@@ -188,13 +190,13 @@ read_keywords(char *s, int *keywords)
 	if (!slen)
 		return 0;
 
-	arch = strtok(s, delim);
-	do {
+	while ((arch = strtok_r(s, delim, &savep)) != NULL) {
+		s = NULL;  /* for strtok_r */
 		i = decode_arch(arch);
 		if (i == -1)
 			continue;
 		keywords[i] = decode_status(arch[0]);
-	} while ((arch = strtok(NULL, delim)));
+	}
 
 	return 0;
 }
@@ -643,7 +645,6 @@ qkeyword_results_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 	qkeyword_data *data = (qkeyword_data *)priv;
 	char buf[_Q_PATH_MAX];
 	depend_atom *patom = NULL;
-	tree_pkg_meta *meta;
 	tree_metadata_xml *metadata;
 	struct elist *emailw;
 	int ret;
@@ -679,11 +680,8 @@ qkeyword_results_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 	}
 
 	keywords = data->keywordsbuf;
-	meta = tree_pkg_read(pkg_ctx);
-	if (meta == NULL)
-		return EXIT_FAILURE;
 
-	if (read_keywords(meta->KEYWORDS, keywords) < 0) {
+	if (read_keywords(tree_pkg_meta_get(pkg_ctx, KEYWORDS), keywords) < 0) {
 		if (verbose)
 			warn("Failed to read keywords for %s%s/%s%s%s",
 				BOLD, pkg_ctx->cat_ctx->name, BLUE, pkg_ctx->name, NORM);
@@ -803,7 +801,8 @@ qkeyword_traverse(tree_pkg_cb func, void *priv)
 	array_for_each(overlays, n, overlay) {
 		tree_ctx *t = tree_open(portroot, overlay);
 		if (t != NULL) {
-			ret |= tree_foreach_pkg_sorted(t, qkeyword_results_cb, priv);
+			ret |= tree_foreach_pkg_sorted(t,
+					qkeyword_results_cb, priv, data->qatom);
 			tree_close(t);
 		}
 	}
