@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2019 Gentoo Foundation
+ * Copyright 2005-2020 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005 Martin Schlemmer     - <azarah@gentoo.org>
@@ -19,6 +19,7 @@
 #include "atom.h"
 #include "contents.h"
 #include "tree.h"
+#include "xpak.h"
 #include "xregex.h"
 
 #define QLIST_FLAGS "IkSRUcDedosF:" COMMON_FLAGS
@@ -103,8 +104,6 @@ umapstr(char display, tree_pkg_ctx *pkg_ctx)
 	char *bufp = _umapstr_buf;
 	char *use = NULL;
 	char *iuse = NULL;
-	size_t use_len;
-	size_t iuse_len;
 	int use_argc = 0;
 	int iuse_argc = 0;
 	char **use_argv = NULL;
@@ -117,11 +116,11 @@ umapstr(char display, tree_pkg_ctx *pkg_ctx)
 	if (!display)
 		return bufp;
 
-	tree_pkg_vdb_eat(pkg_ctx, "USE", &use, &use_len);
-	if (!use[0])
+	use = tree_pkg_meta_get(pkg_ctx, USE);
+	if (use == NULL || *use == '\0')
 		return bufp;
-	tree_pkg_vdb_eat(pkg_ctx, "IUSE", &iuse, &iuse_len);
-	if (!iuse[0])
+	iuse = tree_pkg_meta_get(pkg_ctx, IUSE);
+	if (iuse == NULL || *iuse == '\0')
 		return bufp;
 
 	/* strip out possible leading +/- flags in IUSE */
@@ -166,8 +165,6 @@ umapstr(char display, tree_pkg_ctx *pkg_ctx)
 
 	freeargv(iuse_argc, iuse_argv);
 	freeargv(use_argc, use_argv);
-	free(iuse);
-	free(use);
 
 	return _umapstr_buf;
 }
@@ -261,10 +258,9 @@ qlist_match(
 	if (exact) {
 		int i;
 
-		snprintf(buf, sizeof(buf), "%s/%s-%s%s%s%s%s%s%s",
+		snprintf(buf, sizeof(buf), "%s/%s%s%s%s%s%s%s",
 			atom->CATEGORY,
-			atom->PN,
-			atom->PVR,
+			atom->PF,
 			atom->SLOT != NULL ? ":" : "",
 			atom->SLOT != NULL ? atom->SLOT : "",
 			atom->SUBSLOT != NULL ? "/" : "",
@@ -272,10 +268,10 @@ qlist_match(
 			atom->REPO != NULL ? "::" : "",
 			atom->REPO != NULL ? atom->REPO : "");
 
-		/* exact match: CAT/PN-PVR[:SLOT][::REPO] */
+		/* exact match: CAT/PF[:SLOT][::REPO] */
 		if (strcmp(name, buf) == 0)
 			return true;
-		/* exact match: PN-PVR[:SLOT][::REPO] */
+		/* exact match: PF[:SLOT][::REPO] */
 		if (strcmp(name, strstr(buf, "/") + 1) == 0)
 			return true;
 
@@ -293,15 +289,14 @@ qlist_match(
 		if (strcmp(name, strstr(buf, "/") + 1) == 0)
 			return true;
 	} else {
-		/* partial leading match: CAT/PN-PVR */
-		snprintf(buf, sizeof(buf), "%s/%s-%s",
-				atom->CATEGORY, atom->PN, atom->PVR);
+		/* partial leading match: CAT/PF */
+		snprintf(buf, sizeof(buf), "%s/%s",
+				atom->CATEGORY, atom->PF);
 		if (strncmp(name, buf, pf_len) == 0 ||
 				rematch(name, buf, REG_EXTENDED) == 0)
 			return true;
-		/* partial leading match: PN-PVR */
-		snprintf(buf, sizeof(buf), "%s-%s", atom->PN, atom->PVR);
-		if (strncmp(name, buf, pf_len) == 0 ||
+		/* partial leading match: PF */
+		if (strncmp(name, atom->PF, pf_len) == 0 ||
 				rematch(name, buf, REG_EXTENDED) == 0)
 			return true;
 	}
@@ -315,6 +310,7 @@ struct qlist_opt_state {
 	depend_atom **atoms;
 	bool exact:1;
 	bool all:1;
+	bool do_binpkgs:1;
 	bool just_pkgname:1;
 	bool show_dir:1;
 	bool show_obj:1;
@@ -332,7 +328,9 @@ qlist_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 {
 	struct qlist_opt_state *state = priv;
 	int i;
-	FILE *fp;
+	char *contents;
+	char *line;
+	char *savep;
 	depend_atom *atom;
 
 	/* see if this cat/pkg is requested */
@@ -358,14 +356,15 @@ qlist_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 		printf("%s %sCONTENTS%s:\n",
 				atom_format(state->fmt, atom), DKBLUE, NORM);
 
-	fp = tree_pkg_vdb_fopenat_ro(pkg_ctx, "CONTENTS");
-	if (fp == NULL)
+	if ((contents = tree_pkg_meta_get(pkg_ctx, CONTENTS)) == NULL)
 		return 1;
 
-	while (getline(&state->buf, &state->buflen, fp) != -1) {
+	while ((line = strtok_r(contents, "\n", &savep)) != NULL) {
 		contents_entry *e;
 
-		e = contents_parse_line(state->buf);
+		contents = NULL;  /* for strtok_r */
+
+		e = contents_parse_line(line);
 		if (!e)
 			continue;
 
@@ -397,7 +396,6 @@ qlist_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 				break;
 		}
 	}
-	fclose(fp);
 
 	return 1;
 }
@@ -410,13 +408,13 @@ int qlist_main(int argc, char **argv)
 	int show_slots = 0;
 	bool show_repo = false;
 	bool do_columns = false;
-	bool do_binpkgs = false;
 	char qfmt[128];
 	struct qlist_opt_state state = {
 		.argc = argc,
 		.argv = argv,
 		.exact = false,
 		.all = false,
+		.do_binpkgs = false,
 		.just_pkgname = false,
 		.show_dir = false,
 		.show_obj = false,
@@ -432,7 +430,7 @@ int qlist_main(int argc, char **argv)
 		switch (i) {
 		COMMON_GETOPTS_CASES(qlist)
 		case 'I': state.just_pkgname = true;                    break;
-		case 'k': do_binpkgs = true;                            break;
+		case 'k': state.do_binpkgs = true;                      break;
 		case 'S': state.just_pkgname = true; show_slots++;      break;
 		case 'R': state.just_pkgname = show_repo = true;        break;
 		case 'U': state.just_pkgname = state.show_umap = true;  break;
@@ -488,12 +486,12 @@ int qlist_main(int argc, char **argv)
 	state.buf = xmalloc(state.buflen);
 	state.atoms = xcalloc(argc - optind, sizeof(*state.atoms));
 	ret = 1;
-	if (do_binpkgs)
+	if (state.do_binpkgs)
 		vdb = tree_open_binpkg(portroot, pkgdir);
 	else
 		vdb = tree_open_vdb(portroot, portvdb);
 	if (vdb != NULL) {
-		ret = tree_foreach_pkg_sorted(vdb, qlist_cb, &state);
+		ret = tree_foreach_pkg_sorted(vdb, qlist_cb, &state, NULL);
 		tree_close(vdb);
 	}
 	free(state.buf);
