@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2019 Gentoo Authors
+ * Copyright 2005-2020 Gentoo Authors
  * Distributed under the terms of the GNU General Public License v2
  *
  * Copyright 2005-2010 Ned Ludd        - <solar@gentoo.org>
@@ -272,37 +272,15 @@ qmerge_initialize(void)
 	free(buf);
 }
 
-struct qmerge_bv_state {
-	const char *catname;
-	const char *pkgname;
-	const char *slot;
-	char buf[4096];
-	char *retbuf;
-};
-
-static int
-qmerge_filter_cat(tree_cat_ctx *cat_ctx, void *priv)
-{
-	struct qmerge_bv_state *state = priv;
-	return !state->catname || strcmp(cat_ctx->name, state->catname) == 0;
-}
-
-/* HACK: pull this in, knowing that qlist will be in the final link, we
- * should however figure out how to do what match does here from e.g.
- * atom */
-extern bool qlist_match(
-		tree_pkg_ctx *pkg_ctx,
-		const char *name,
-		depend_atom **name_atom,
-		bool exact);
-
+static char _best_version_retbuf[4096];
 static int
 qmerge_best_version_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 {
-	struct qmerge_bv_state *state = priv;
-	if (qlist_match(pkg_ctx, state->buf, NULL, true))
-		snprintf(state->retbuf, sizeof(state->buf), "%s/%s:%s",
-			 pkg_ctx->cat_ctx->name, pkg_ctx->name, state->slot);
+	depend_atom *sa = priv;
+	depend_atom *a = tree_get_atom(pkg_ctx, true);  /* need SLOT */
+	if (atom_compare(a, sa) == EQUAL)
+		snprintf(_best_version_retbuf, sizeof(_best_version_retbuf),
+				"%s/%s:%s", a->CATEGORY, a->PF, a->SLOT);
 	return 0;
 }
 
@@ -310,15 +288,7 @@ static char *
 best_version(const char *catname, const char *pkgname, const char *slot)
 {
 	static int vdb_check = 1;
-	static char retbuf[4096];
-
 	tree_ctx *vdb;
-	struct qmerge_bv_state state = {
-		.catname = catname,
-		.pkgname = pkgname,
-		.slot = slot,
-		.retbuf = retbuf,
-	};
 
 	/* Make sure these dirs exist before we try walking them */
 	switch (vdb_check) {
@@ -336,18 +306,18 @@ best_version(const char *catname, const char *pkgname, const char *slot)
 			goto done;
 	}
 
-	retbuf[0] = '\0';
-	snprintf(state.buf, sizeof(state.buf), "%s%s%s:%s",
-		 catname ? : "", catname ? "/" : "", pkgname, slot);
+	snprintf(_best_version_retbuf, sizeof(_best_version_retbuf),
+			"%s%s%s:%s", catname ? : "", catname ? "/" : "", pkgname, slot);
 	vdb = tree_open_vdb(portroot, portvdb);
 	if (vdb != NULL) {
-		tree_foreach_pkg_fast(vdb,
-				qmerge_best_version_cb, &state, qmerge_filter_cat);
+		depend_atom *sa = atom_explode(_best_version_retbuf);
+		tree_foreach_pkg_fast(vdb, qmerge_best_version_cb, sa, sa);
 		tree_close(vdb);
+		atom_implode(sa);
 	}
 
  done:
-	return retbuf;
+	return _best_version_retbuf;
 }
 
 static int
@@ -603,15 +573,15 @@ install_mask_pwd(int iargc, char **iargv, const struct stat * const st, int fd)
 	}
 
 #if EBUG
-	printf("applying install masks:\n");
+	fprintf(warnout, "applying install masks:\n");
 	for (cnt = 0; cnt < masksc; cnt++) {
 		ssize_t plen = (ssize_t)masksv[cnt][0];
-		printf("%3zd  ", plen);
+		fprintf(warnout, "%3zd  ", plen);
 		if (plen < 0)
 			plen = -plen;
 		for (i = 1; i <= plen; i++)
-			printf("%s ", masksv[cnt][i]);
-		printf(" %zd\n", (size_t)masksv[cnt][i]);
+			fprintf(warnout, "%s ", masksv[cnt][i]);
+		fprintf(warnout, " %zd\n", (size_t)masksv[cnt][i]);
 	}
 #endif
 
@@ -1000,6 +970,7 @@ merge_tree_at(int fd_src, const char *src, int fd_dst, const char *dst,
 		}
 	}
 
+	closedir(dir);
 	ret = 0;
 
  done:
@@ -1454,19 +1425,16 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, set *keep,
 		int cp_argc, char **cp_argv, int cpm_argc, char **cpm_argv)
 {
 	tree_cat_ctx *cat_ctx = pkg_ctx->cat_ctx;
-	size_t buflen;
-	static char *phases;
-	static size_t phases_len;
-	char *eprefix = NULL;
-	size_t eprefix_len = 0;
+	char *phases;
+	char *eprefix;
+	size_t eprefix_len;
 	const char *T;
 	char *buf;
-	FILE *fp;
-	int ret, portroot_fd;
+	char *savep;
+	int portroot_fd;
 	llist_char *dirs = NULL;
 	bool unmerge_config_protected;
 
-	ret = 1;
 	buf = phases = NULL;
 	T = "${PWD}/temp";
 
@@ -1477,28 +1445,31 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, set *keep,
 		return 0;
 
 	/* First get a handle on the things to clean up */
-	fp = tree_pkg_vdb_fopenat_ro(pkg_ctx, "CONTENTS");
-	if (fp == NULL)
-		return ret;
+	buf = tree_pkg_meta_get(pkg_ctx, CONTENTS);
+	if (buf == NULL)
+		return 1;
 
 	portroot_fd = cat_ctx->ctx->portroot_fd;
 
 	/* Then execute the pkg_prerm step */
 	if (!pretend) {
-		tree_pkg_vdb_eat(pkg_ctx, "DEFINED_PHASES", &phases, &phases_len);
-		mkdirat(pkg_ctx->fd, "temp", 0755);
-		pkg_run_func_at(pkg_ctx->fd, ".", phases, "pkg_prerm", T, T);
+		phases = tree_pkg_meta_get(pkg_ctx, DEFINED_PHASES);
+		if (phases != NULL) {
+			mkdirat(pkg_ctx->fd, "temp", 0755);
+			pkg_run_func_at(pkg_ctx->fd, ".", phases, "pkg_prerm", T, T);
+		}
 	}
 
-	if (!tree_pkg_vdb_eat(pkg_ctx, "EPREFIX", &eprefix, &eprefix_len))
+	eprefix = tree_pkg_meta_get(pkg_ctx, EPREFIX);
+	if (eprefix == NULL)
 		eprefix_len = 0;
-	if (eprefix != NULL)
-		free(eprefix);
+	else
+		eprefix_len = strlen(eprefix);
 
 	unmerge_config_protected =
 		strstr(features, "config-protect-if-modified") != NULL;
 
-	while (getline(&buf, &buflen, fp) != -1) {
+	for (; (buf = strtok_r(buf, "\n", &savep)) != NULL; buf = NULL) {
 		bool del;
 		contents_entry *e;
 		char zing[20];
@@ -1596,8 +1567,6 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, set *keep,
 		}
 	}
 
-	fclose(fp);
-
 	/* Then remove all dirs in reverse order */
 	while (dirs != NULL) {
 		llist_char *list = dirs;
@@ -1625,11 +1594,7 @@ pkg_unmerge(tree_pkg_ctx *pkg_ctx, set *keep,
 		unlinkat(cat_ctx->ctx->tree_fd, cat_ctx->name, AT_REMOVEDIR);
 	}
 
-	ret = 0;
-	free(phases);
-	free(buf);
-
-	return ret;
+	return 0;
 }
 
 static int
@@ -1810,6 +1775,15 @@ print_Pkg(int full, const depend_atom *atom, const struct pkg_t *pkg)
 	}
 }
 
+/* HACK: pull this in, knowing that qlist will be in the final link, we
+ * should however figure out how to do what match does here from e.g.
+ * atom */
+extern bool qlist_match(
+		tree_pkg_ctx *pkg_ctx,
+		const char *name,
+		depend_atom **name_atom,
+		bool exact);
+
 static int
 qmerge_unmerge_cb(tree_pkg_ctx *pkg_ctx, void *priv)
 {
